from labrea.template import Template
from labrea.exceptions import KeyNotFoundError
import pytest


def test_basic():
    template = Template('{GREETING}, {:name:}!', name='World')
    options = {'GREETING': 'Hello'}

    assert template(options) == 'Hello, World!'
    template.validate(options)
    assert template.keys(options) == {'GREETING'}
    assert template.explain(options) == {'GREETING'}


def test_malformed():
    with pytest.raises(ValueError):
        Template('{GREETING} {:name:}')

    with pytest.warns(UserWarning):
        Template('{GREETING}', name='World')


def test_missing():
    template = Template('{GREETING}, {:name:}!', name='World')
    options = {}

    with pytest.raises(KeyNotFoundError) as excinfo:
        template(options)
        assert excinfo.value.key == 'GREETING'

    with pytest.raises(KeyNotFoundError) as excinfo:
        template.validate(options)
        assert excinfo.value.key == 'GREETING'

    with pytest.raises(KeyNotFoundError) as excinfo:
        template.keys(options)
        assert excinfo.value.key == 'GREETING'

    assert template.explain(options) == {'GREETING'}


def test_confectioner_templating():
    template = Template('{A} {B}{:punc:}', punc='!')
    present = {'A': '{C}', 'B': '{D}', 'C': 'Hello', 'D': 'World'}
    missing = {'A': '{C}', 'B': '{D}', 'C': 'Hello'}

    assert template(present) == 'Hello World!'
    template.validate(present)
    assert template.keys(present) == {'A', 'B', 'C', 'D'}

    with pytest.raises(KeyNotFoundError) as excinfo:
        template(missing)
        assert excinfo.value.key == 'D'

    with pytest.raises(KeyNotFoundError) as excinfo:
        template.validate(missing)
        assert excinfo.value.key == 'D'

    with pytest.raises(KeyNotFoundError) as excinfo:
        template.keys(missing)
        assert excinfo.value.key == 'D'

    assert template.explain(present) == {'A', 'B', 'C', 'D'}
    assert template.explain(missing) == {'A', 'B', 'C', 'D'}
    assert template.explain() == {'A', 'B'}


def test_repr():
    assert repr(Template('{GREETING}, {:name:}!', name='World')) == "Template('{GREETING}, {:name:}!', name=Value('World'))"
    assert repr(Template('{GREETING}')) == "Template('{GREETING}')"



import re
import warnings
from typing import Any, Dict, Optional, Set

from confectioner import mix
from confectioner.templating import find_template_keys, resolve

from .exceptions import KeyNotFoundError
from .types import Evaluatable, Options, Value

TEMPLATE_PARAM = re.compile(r"^:[a-zA-Z_][a-zA-Z0-9_]*:$")


class Template(Evaluatable[str]):
    """A template string that can be evaluated using other Evaluatables.

    Template entries use the format :code:`{:key:}` where :code:`key` is the
    name of a parameter (keyword argument). The parameter can be any
    Evaluatable. The template is evaluated using the options dictionary and
    the result is returned as a string. Normal options can also be used in
    the template using the standard :code:`{NESTED.KEY}` syntax from
    confectioner.

    Arguments
    ----------
    template : str
        The template string to evaluate
    kwargs : Any
        The parameters to use when evaluating the template. Each parameter
        can be any Evaluatable. If a parameter is not an Evaluatable, it is
        used as a constant key.

    Raises
    ------
    ValueError
        If the template requires parameters that are not provided


    Example Usage
    -------------
    >>> from labrea import Template, dataset, Option
    >>> @dataset
    ... def b_dataset(b: str = Option('B')) -> str:
    ...     return b
    >>>
    >>> t = Template('{A.X} {:b:}', b=b_dataset)
    >>> t({'A': {'X': 'Hello'}, 'B': 'World!'})
    'Hello World!'
    """

    template: str
    params: Dict[str, Evaluatable[Any]]

    def __init__(self, template: str, **kwargs):
        self.template = template
        self.params = {
            key: value if isinstance(value, Evaluatable) else Value(value)
            for key, value in kwargs.items()
        }

        required_params = {
            key[1:-1]
            for key in find_template_keys(template)
            if TEMPLATE_PARAM.match(key)
        }

        missing_params = required_params - self.params.keys()
        if missing_params:
            raise ValueError(
                f"Template {template} requires parameters "
                f'{",".join(required_params)}'
            )

        extra_params = set(self.params.keys()) - required_params
        if extra_params:
            warnings.warn(
                f'Unused parameters {",".join(extra_params)} for template '
                f"{template}"
            )

    def evaluate(self, options: Options) -> str:
        """Evaluates the template using the options dictionary."""
        params = {f":{key}:": val.evaluate(options) for key, val in self.params.items()}

        try:
            return str(resolve(self.template, mix(options, params)))  # type: ignore
        except KeyError as e:
            raise KeyNotFoundError((*e.args, "UNKNOWN")[0], self) from e

    def validate(self, options: Options) -> None:
        """Validates that the template can be evaluated using the options."""
        from .option import Option

        for val in self.params.values():
            val.validate(options)

        for key in find_template_keys(self.template):
            if TEMPLATE_PARAM.match(key):
                continue
            try:
                Option(key).validate(options)
            except KeyNotFoundError as e:
                raise KeyNotFoundError(e.key, self) from e

    def keys(self, options: Options) -> Set[str]:
        """Returns the keys that this object depends on."""
        from .option import Option

        keys = set().union(*(value.keys(options) for value in self.params.values()))
        for key in find_template_keys(self.template):
            if TEMPLATE_PARAM.match(key):
                continue
            try:
                keys.update(Option(key).keys(options))
            except KeyNotFoundError as e:
                raise KeyNotFoundError(e.key, self) from e

        return keys

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Returns the keys that this object depends on."""
        from .option import Option

        options = options or {}
        keys = set().union(*(value.explain(options) for value in self.params.values()))
        for key in find_template_keys(self.template):
            if TEMPLATE_PARAM.match(key):
                continue

            keys.update(Option(key).explain(options))

        return keys

    def __repr__(self) -> str:
        if self.params:
            return (
                f"Template({self.template!r}, "
                f"{', '.join(f'{key}={value!r}' for key, value in self.params.items())}"
                f")"
            )

        return f"Template({self.template!r})"



import sys

if sys.version_info < (3, 11):
    from typing import NoReturn as Never
else:
    from typing import Never

import functools
from typing import (
    Any,
    Callable,
    Generic,
    Hashable,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from ._missing import MISSING, MaybeMissing
from .exceptions import EvaluationError, InsufficientInformationError
from .option import Option
from .types import Evaluatable, MaybeEvaluatable, Options

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
V = TypeVar("V")


class SwitchError(EvaluationError):
    """An error raised when a switch statement encounters an invalid key."""

    def __init__(
        self,
        dispatch: Evaluatable[Hashable],
        value: Hashable,
        lookup: Mapping[Hashable, Any],
    ):
        super().__init__(
            f"Evaluated to {value}, "
            f"but must be one of {', '.join(map(str, lookup.keys()))}.",
            dispatch,
        )


class _DependsOn(Generic[A, B], Evaluatable[B]):
    evaluatable: Evaluatable[B]
    depends: Evaluatable[A]

    def __init__(
        self, evaluatable: MaybeEvaluatable[B], depends: MaybeEvaluatable[A]
    ) -> None:
        self.evaluatable = Evaluatable.ensure(evaluatable)
        self.depends = Evaluatable.ensure(depends)

    def evaluate(self, options: Options) -> B:
        return self.evaluatable.evaluate(options)

    def validate(self, options: Options) -> None:
        self.evaluatable.validate(options)

    def keys(self, options: Options) -> Set[str]:
        return self.evaluatable.keys(options) | self.depends.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return self.evaluatable.explain(options) | self.depends.explain(options)

    def __repr__(self) -> str:
        return f"_DependsOn({self.evaluatable!r}, {self.depends!r})"  # pragma: no cover


class Switch(Evaluatable[V]):
    """A class representing a switch statement.

    This class takes a dispatch evaluatable and a mapping of keys to evaluatables.
    When evaluated, the dispatch is evaluated and the corresponding value is
    evaluated and returned. If the key is not found in the mapping, or the dispatch
    cannot be evaluated, a default value can be provided. If no default is provided,
    and the switch cannot choose a branch, an error is raised.

    Aliases: :func:`labrea.switch`, :class:`labrea.Switch`


    Arguments
    ---------
    dispatch : Union[str, Evaluatable[Hashable]]
        The dispatch evaluatable. This is used to determine which branch to take.
    lookup : Mapping[Hashable, MaybeEvaluatable[V]]
        A mapping of keys to evaluatables. The key is determined by the dispatch.
        Values can also be plain values.
    default : MaybeMissing[MaybeEvaluatable[V]], optional
        The default value to use if the key is not found in the mapping. Can be
        an evaluatable, or a plain value.


    Example Usage
    -------------
    >>> from labrea import Option, switch
    >>> s = switch(Option('A'), {True: Option('X'), False: Option('Y')}, Option('Z'))
    >>> s({'A': True, 'X': 'Hello', 'Y': 'World', 'Z': 'Default'})
    'Hello'
    >>> s({'A': False, 'X': 'Hello', 'Y': 'World', 'Z': 'Default'})
    'World'
    >>> s({'A': None, 'X': 'Hello', 'Y': 'World', 'Z': 'Default'})
    'Default'
    >>> s({'X': 'Hello', 'Y': 'World', 'Z': 'Default'})
    'Default'
    """

    dispatch: Evaluatable[Hashable]
    lookup: Mapping[Hashable, Evaluatable[V]]
    default: MaybeMissing[Evaluatable[V]]

    def __init__(
        self,
        dispatch: Union[str, Evaluatable[Hashable]],
        lookup: Mapping[Hashable, Evaluatable[V]],
        default: MaybeMissing[MaybeEvaluatable[V]] = MISSING,
    ) -> None:
        self.dispatch = (
            dispatch if isinstance(dispatch, Evaluatable) else Option(dispatch)
        )
        self.lookup = {key: Evaluatable.ensure(value) for key, value in lookup.items()}
        self.default = (
            Evaluatable.ensure(default) if default is not MISSING else default
        )

    def _dispatch(self, options: Options) -> Hashable:
        return self.dispatch.evaluate(options)

    def _lookup(self, options: Options) -> Evaluatable[V]:
        try:
            key = self._dispatch(options)
        except EvaluationError as e:
            if self.default is MISSING:
                raise e
            return self.default

        if key not in self.lookup:
            if self.default is MISSING:
                raise SwitchError(self.dispatch, key, self.lookup)  # type: ignore  [arg-type]
            return _DependsOn(self.default, self.dispatch)  # type: ignore  [arg-type]

        return _DependsOn(self.lookup[key], self.dispatch)  # type: ignore  [arg-type]

    def evaluate(self, options: Options) -> V:
        """Evaluate the switch statement and return the result."""
        return self._lookup(options).evaluate(options)

    def validate(self, options: Options) -> None:
        """Validate that the switch statement can be evaluated."""
        self._lookup(options).validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required by the switch statement."""
        return self._lookup(options).keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required by the switch statement."""
        options = options or {}
        try:
            chosen = self._lookup(options)
        except EvaluationError as e:
            raise InsufficientInformationError(
                "Could not determine the switch branch", self
            ) from e

        return chosen.explain(options)

    def __repr__(self) -> str:
        if self.default is MISSING:
            return f"switch({self.dispatch!r}, {self.lookup!r})"

        return f"switch({self.dispatch!r}, {self.lookup!r}, {self.default!r})"


switch = Switch


class CaseWhenError(EvaluationError):
    """An error raised when a case when statement encounters an invalid key."""

    def __init__(self, dispatch: Evaluatable[A], value: A):
        super().__init__(
            f"Evaluated to {value}, which does not match any of the cases, "
            f"and no default was provided.",
            dispatch,
        )


class CaseWhen(Generic[A, B], Evaluatable[B]):
    """A class representing a case when statement.

    This class takes a dispatch evaluatable and a sequence of cases. Each case is
    a condition and a result. When evaluated, the dispatch is evaluated and the
    corresponding case is matched. If no case is matched, a default value can be
    provided. If no default is provided, and the case cannot choose a branch, an
    error is raised.

    This class is not usually instantiated directly. Instead, use the :code:`case` function.
    """

    dispatch: Evaluatable[A]
    cases: Sequence[Tuple[Evaluatable[Callable[[A], bool]], Evaluatable[B]]]
    default: MaybeMissing[Evaluatable[B]]

    def __init__(
        self,
        dispatch: Evaluatable[A],
        cases: Sequence[Tuple[Evaluatable[Callable[[A], bool]], Evaluatable[B]]],
        default: MaybeMissing[Evaluatable[B]] = MISSING,
    ) -> None:
        self.dispatch = dispatch
        self.cases = [(condition, result) for condition, result in cases]
        self.default = default

    def _evaluate(self, value: A, options: Options) -> Evaluatable[B]:
        for condition, result in self.cases:
            if condition.evaluate(options)(value):
                return result

        if self.default is not MISSING:
            return self.default

        raise CaseWhenError(self.dispatch, value)

    def _bound(self, options: Options) -> Evaluatable[B]:
        return self.dispatch.bind(functools.partial(self._evaluate, options=options))

    def evaluate(self, options: Options) -> B:
        """Evaluate the case when statement and return the result."""
        return self._bound(options).evaluate(options)

    def validate(self, options: Options) -> None:
        """Validate that the case when statement can be evaluated."""
        self._bound(options).validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required by the case when statement."""
        return self._bound(options).keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required by the case when statement."""
        return self._bound(options or {}).explain(options)

    @overload
    def when(
        self,
        condition: MaybeEvaluatable[Callable[[A], bool]],
        result: Evaluatable[C],
    ) -> "CaseWhen[A, Union[B, C]]": ...

    @overload
    def when(
        self,
        condition: MaybeEvaluatable[Callable[[A], bool]],
        result: C,
    ) -> "CaseWhen[A, Union[B, C]]": ...

    def when(
        self,
        condition: MaybeEvaluatable[Callable[[A], bool]],
        result: MaybeEvaluatable[C],
    ) -> "CaseWhen[A, Union[B, C]]":
        """Add a case to the case when statement. Returns a new instance.

        Arguments
        ---------
        condition : Union[str, Callable[[A], bool]]
            The condition to match. Can be a plain value or a callable.
        result : MaybeEvaluatable[B]
            The result to return if the condition is matched.
        """
        return CaseWhen(
            self.dispatch,
            [*self.cases, (Evaluatable.ensure(condition), Evaluatable.ensure(result))],
            self.default,
        )

    @overload
    def otherwise(self, default: Evaluatable[C]) -> "CaseWhen[A, Union[B, C]]": ...

    @overload
    def otherwise(self, default: C) -> "CaseWhen[A, Union[B, C]]": ...

    def otherwise(self, default: MaybeEvaluatable[C]) -> "CaseWhen[A, Union[B, C]]":
        """Add a default value to the case when statement. Returns a new instance.

        Arguments
        ---------
        default : MaybeEvaluatable[B]
            The default value to return if no case is matched.
        """
        return CaseWhen(self.dispatch, self.cases, Evaluatable.ensure(default))

    def __repr__(self) -> str:
        _base = f"case({self.dispatch!r})"
        _cases = ", ".join(
            f"when({condition!r}, {result!r})" for condition, result in self.cases
        )
        _default = (
            f".otherwise({self.default!r})" if self.default is not MISSING else ""
        )
        return f"{_base}.{_cases}{_default}"


@overload
def case(dispatch: Evaluatable[A]) -> CaseWhen[A, Never]: ...


@overload
def case(dispatch: A) -> CaseWhen[A, Never]: ...


def case(dispatch: MaybeEvaluatable[A]) -> CaseWhen[A, Never]:
    """Create a case when statement.

    This is the preferred method for creating a case when statement.
    It allows for chaining multiple cases and a default value.

    Arguments
    ---------
    dispatch : MaybeEvaluatable[A]
        The dispatch evaluatable. This is used to determine which branch to take.

    Returns
    -------
    CaseWhen[A, B]
        A new case when statement with no cases or default.


    Example Usage
    -------------
    >>> from labrea import Option, case
    >>> c = case(
    ...     Option('A')
    ... ).when(
    ...     lambda a: a < 0,
    ...     Option('X')
    ... ).when(
    ...     lambda a: a > 0,
    ...     Option('Y')
    ... ).otherwise(Option('Z'))
    >>> c({'A': -1, 'X': 'Negative', 'Y': 'Positive', 'Z': 'Zero'})
    'Negative'
    >>> c({'A': 1, 'X': 'Negative', 'Y': 'Positive', 'Z': 'Zero'})
    'Positive'
    >>> c({'A': 0, 'X': 'Negative', 'Y': 'Positive', 'Z': 'Zero'})
    'Zero'
    """
    return CaseWhen(Evaluatable.ensure(dispatch), [])



from typing import Dict, Hashable, List, Set, Tuple, TypeVar, Union, overload

from .iterable import Iter
from .types import Evaluatable, Value

A = TypeVar("A")
K = TypeVar("K", bound=Hashable)
V = TypeVar("V")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")


def evaluatable_list(*evaluatables: Evaluatable[A]) -> Evaluatable[List[A]]:
    """Create an Evaluatable that evaluates to a list of values.

    This function is a convenience wrapper around Iter.apply(list).

    Aliases: :func:`labrea.DatasetList`, :func:`labrea.evaluatable_list`

    Arguments
    ---------
    *evaluatables : Evaluatable[A]
        The objects to evaluate.
    """
    return Iter(*evaluatables).apply(list)


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
) -> Evaluatable[Tuple[T1]]: ...


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
    __e2: Evaluatable[T2],
) -> Evaluatable[Tuple[T1, T2]]: ...


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
    __e2: Evaluatable[T2],
    __e3: Evaluatable[T3],
) -> Evaluatable[Tuple[T1, T2, T3]]: ...


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
    __e2: Evaluatable[T2],
    __e3: Evaluatable[T3],
    __e4: Evaluatable[T4],
) -> Evaluatable[Tuple[T1, T2, T3, T4]]: ...


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
    __e2: Evaluatable[T2],
    __e3: Evaluatable[T3],
    __e4: Evaluatable[T4],
    __e5: Evaluatable[T5],
) -> Evaluatable[Tuple[T1, T2, T3, T4, T5]]: ...


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
    __e2: Evaluatable[T2],
    __e3: Evaluatable[T3],
    __e4: Evaluatable[T4],
    __e5: Evaluatable[T5],
    __e6: Evaluatable[T6],
) -> Evaluatable[Tuple[T1, T2, T3, T4, T5, T6]]: ...


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
    __e2: Evaluatable[T2],
    __e3: Evaluatable[T3],
    __e4: Evaluatable[T4],
    __e5: Evaluatable[T5],
    __e6: Evaluatable[T6],
    __e7: Evaluatable[T7],
) -> Evaluatable[Tuple[T1, T2, T3, T4, T5, T6, T7]]: ...


@overload
def evaluatable_tuple(
    __e1: Evaluatable[T1],
    __e2: Evaluatable[T2],
    __e3: Evaluatable[T3],
    __e4: Evaluatable[T4],
    __e5: Evaluatable[T5],
    __e6: Evaluatable[T6],
    __e7: Evaluatable[T7],
    __e8: Evaluatable[T8],
) -> Evaluatable[Tuple[T1, T2, T3, T4, T5, T6, T7, T8]]: ...


def evaluatable_tuple(*evaluatables: Evaluatable[A]) -> Evaluatable[Tuple[A, ...]]:
    """Create an Evaluatable that evaluates to a tuple of values.

    This function is a convenience wrapper around Iter.apply(tuple).

    Aliases: :func:`labrea.DatasetTuple`, :func:`labrea.evaluatable_tuple`

    Arguments
    ---------
    *evaluatables : Evaluatable[A]
        The objects to evaluate.
    """
    return Iter(*evaluatables).apply(tuple)


def evaluatable_set(*evaluatables: Evaluatable[K]) -> Evaluatable[Set[K]]:
    """Create an Evaluatable that evaluates to a set of values.

    This function is a convenience wrapper around Iter.apply(set).

    Aliases: :func:`labrea.DatasetSet`, :func:`labrea.evaluatable_set`

    Arguments
    ---------
    *evaluatables : Evaluatable[K]
        The objects to evaluate.
    """
    return Iter(*evaluatables).apply(set)


def evaluatable_dict(contents: Dict[K, Evaluatable[V]]) -> Evaluatable[Dict[K, V]]:
    """Create an Evaluatable that evaluates to a dictionary of values.

    This function is a convenience wrapper around Iter.apply(dict).

    Aliases: :func:`labrea.DatasetDict`, :func:`labrea.evaluatable_dict`

    Arguments
    ---------
    contents : Dict[K, Evaluatable[V]]
        The objects to evaluate.
    """
    pairs = (Iter[Union[K, V]](Value(key), val) for key, val in contents.items())
    return Iter(*pairs).apply(dict)  # type: ignore


DatasetList = evaluatable_list
DatasetTuple = evaluatable_tuple
DatasetSet = evaluatable_set
DatasetDict = evaluatable_dict



from enum import Enum
from typing import TypeVar, Union

T = TypeVar("T")


class Missing(Enum):
    token = 0

    def __repr__(self) -> str:
        return "MISSING"


MISSING = Missing.token


MaybeMissing = Union[Missing, T]



import logging
import pytest
import uuid
import pickle

from labrea.computation import CallbackEffect
from labrea.dataset import Dataset, dataset, abstractdataset
from labrea.exceptions import EvaluationError
from labrea.logging import LogRequest
from labrea.option import Option
import labrea.runtime


def test_dataset():
    @dataset
    def add(a: int = Option('A'), b: int = Option('B')) -> int:
        return a + b

    assert add.evaluate({'A': 1, 'B': 2}) == 3
    add.validate({'A': 1, 'B': 2})
    assert add.keys({'A': 1, 'B': 2}) == {'A', 'B'}
    assert add.explain() == {'A', 'B'}


def test_dataset_non_decorator() -> None:
    a = dataset(Option('A'))
    assert isinstance(a, Dataset)
    assert a.evaluate({'A': 1}) == 1


def test_overloads():
    @dataset(dispatch='X')
    def combine(a: int = Option('A'), b: int = Option('B')) -> int:
        return a + b

    @combine.overload(alias='Y')
    def multiply(a: int = Option('C'), b: int = Option('D')) -> int:
        return a * b

    assert combine.evaluate({'A': 1, 'B': 2}) == 3
    combine.validate({'A': 1, 'B': 2})
    assert combine.keys({'A': 1, 'B': 2}) == {'A', 'B'}
    assert combine.explain() == {'A', 'B'}

    assert multiply.evaluate({'C': 1, 'D': 2}) == 2
    multiply.validate({'C': 1, 'D': 2})
    assert multiply.keys({'C': 1, 'D': 2}) == {'C', 'D'}
    assert multiply.explain() == {'C', 'D'}

    assert combine.evaluate({'X': 'Y', 'C': 1, 'D': 2}) == 2
    combine.validate({'X': 'Y', 'C': 1, 'D': 2})
    assert combine.keys({'X': 'Y', 'C': 1, 'D': 2}) == {'X', 'C', 'D'}
    assert combine.explain({'X': 'Y'}) == {'X', 'C', 'D'}

    assert combine.evaluate({'X': 'Z', 'A': 1, 'B': 2}) == 3
    combine.validate({'X': 'Z', 'A': 1, 'B': 2})
    assert combine.keys({'X': 'Z', 'A': 1, 'B': 2}) == {'X', 'A', 'B'}
    assert combine.explain({'X': 'Z'}) == {'X', 'A', 'B'}


def test_overload_no_dispatch():
    @dataset
    def x():
        pass

    with pytest.raises(ValueError):
        @x.overload('A')
        def y():
            pass


def test_abstract():

    @abstractdataset(dispatch='A')
    def x() -> int:
        pass

    @x.overload('B')
    def y() -> int:
        return 1

    with pytest.raises(EvaluationError):
        x()

    assert x.evaluate({'A': 'B'}) == 1


def test_repr():
    def add(a: int = Option('A'), b: int = Option('B')) -> int:
        return a + b

    assert repr(dataset(add)) == f"<Dataset {add.__qualname__}>"


def test_cache():
    @dataset
    def uuid4() -> uuid.UUID:
        return uuid.uuid4()

    assert uuid4() == uuid4()


def test_nocache():
    @dataset.nocache
    def uuid4() -> uuid.UUID:
        return uuid.uuid4()

    assert uuid4() != uuid4()


def test_effect():
    store = None

    def set_store(value):
        nonlocal store
        store = value
        return value

    @dataset.nocache(effects=[CallbackEffect(set_store)])
    def uuid4() -> uuid.UUID:
        return uuid.uuid4()

    result = uuid4()
    assert store == result

    @dataset.nocache(effects=[set_store])
    def uuid4() -> uuid.UUID:
        return uuid.uuid4()

    result = uuid4()
    assert store == result

    @dataset.nocache
    def uuid4() -> uuid.UUID:
        return uuid.uuid4()

    uuid4.add_effect(CallbackEffect(set_store))
    result = uuid4()
    assert store == result

    result = uuid4()
    assert store == result

    @dataset.nocache(effects=[set_store])
    def uuid4() -> uuid.UUID:
        return uuid.uuid4()

    result1 = uuid4()
    assert store == result1

    uuid4.disable_effects()
    result2 = uuid4()
    assert store != result2

    uuid4.enable_effects()
    result3 = uuid4()
    assert store == result3

    result4 = uuid4({'LABREA': {'EFFECTS': {'DISABLED': True}}})
    assert store != result4


def test_set_dispatch():
    @dataset
    def x() -> int:
        return 1

    x.set_dispatch(Option('A'))

    @x.overload('A')
    def y() -> int:
        return 2

    assert x() == 1
    assert x.evaluate({'A': 'A'}) == 2


def test_where():
    def add(x: int, y: int) -> int:
        return x + y

    add = dataset.where(x=Option('X'), y=Option('Y')).wrap(add)

    assert add.evaluate({'X': 1, 'Y': 2}) == 3
    add.validate({'X': 1, 'Y': 2})
    assert add.keys({'X': 1, 'Y': 2}) == {'X', 'Y'}
    assert add.explain() == {'X', 'Y'}


def test_force_options():
    @dataset(options={'A': 1})
    def x(a: int = Option('A'), b: int = Option('B')) -> int:
        return a + b

    assert x({'B': 2}) == x({'A': 2, 'B': 2}) == 3
    x.validate({'B': 2})
    assert x.keys({'B': 2}) == {'B'}
    assert x.explain() == {'B'}

    with pytest.raises(EvaluationError):
        x()
    with pytest.raises(EvaluationError):
        x.validate({})
    with pytest.raises(EvaluationError):
        x.keys({})


def test_default_options():
    @dataset(default_options={'A': 1})
    def x(a: int = Option('A'), b: int = Option('B')) -> int:
        return a + b

    assert x({'B': 2}) == 3
    x.validate({'B': 2})
    assert x.keys({'B': 2}) == {'B'}
    assert x.explain() == {'B'}

    assert x({'A': 2, 'B': 2}) == 4
    x.validate({'A': 2, 'B': 2})
    assert x.keys({'A': 2, 'B': 2}) == {'A', 'B'}
    assert x.explain({'A': 2}) == {'A', 'B'}

    with pytest.raises(EvaluationError):
        x()
    with pytest.raises(EvaluationError):
        x.validate({})
    with pytest.raises(EvaluationError):
        x.keys({})


def test_default():
    @dataset(dispatch='X')
    def x() -> str:
        return 'x'

    @x.overload('Y')
    def y() -> str:
        return 'y'

    assert x.default({'X': 'Y'}) == 'x'


def test_with_options():
    @dataset
    def x(a: int = Option('A'), b: int = Option('B')) -> int:
        return a + b

    x1 = x.with_options({'A': 1})
    x2 = x.with_options({'A': 2})

    assert isinstance(x1, type(x))

    assert x1({'B': 2}) == 3
    assert x2({'B': 2}) == 4


def test_with_default_options():
    @dataset
    def x(a: int = Option('A'), b: int = Option('B')) -> int:
        return a + b

    x1 = x.with_default_options({'A': 1})

    assert isinstance(x1, type(x))

    assert x1({'B': 2}) == 3
    assert x1({'A': 2, 'B': 2}) == 4


def test_logging():
    @dataset
    def x() -> int:
        return 1

    handler_run = False

    def test_logging_handler(request: LogRequest) -> None:
        nonlocal handler_run
        handler_run = True
        assert request.level == logging.INFO
        assert request.name == x.__module__
        assert request.msg == f'Labrea: Evaluating {x!r}'

    with labrea.runtime.handle(LogRequest, test_logging_handler):
        x()

    assert handler_run


def test_pickle():
    x = dataset(Option('X'))

    assert isinstance(pickle.loads(pickle.dumps(x)), Dataset)



from labrea.conditional import switch, SwitchError, case, CaseWhenError
from labrea.exceptions import KeyNotFoundError, InsufficientInformationError
from labrea.option import Option
import pytest


def test_switch():
    s = switch(Option('A'), {'X': 42, 'Y': Option('Z')})

    assert s({'A': 'X'}) == 42
    assert s({'A': 'Y', 'Z': 43}) == 43
    with pytest.raises(KeyNotFoundError):
        s.evaluate({})
    with pytest.raises(KeyNotFoundError):
        s.evaluate({'A': 'Y'})
    with pytest.raises(SwitchError):
        s.evaluate({'A': 'Z'})

    s.validate({'A': 'X'})
    s.validate({'A': 'Y', 'Z': 43})
    with pytest.raises(KeyNotFoundError):
        s.validate({})
    with pytest.raises(KeyNotFoundError):
        s.validate({'A': 'Y'})
    with pytest.raises(SwitchError):
        s.validate({'A': 'Z'})

    assert s.keys({'A': 'X'}) == {'A'}
    assert s.keys({'A': 'Y', 'Z': 43}) == {'A', 'Z'}
    with pytest.raises(KeyNotFoundError):
        s.keys({})
    with pytest.raises(KeyNotFoundError):
        s.keys({'A': 'Y'})
    with pytest.raises(SwitchError):
        s.keys({'A': 'Z'})

    assert s.explain({'A': 'X'}) == {'A'}
    assert s.explain({'A': 'Y', 'Z': 43}) == {'A', 'Z'}
    with pytest.raises(InsufficientInformationError):
        s.explain({})
    assert s.explain({'A': 'Y'}) == {'A', 'Z'}
    with pytest.raises(InsufficientInformationError):
        assert s.explain({'A': 'Z'}) == {'A'}

    assert repr(s) == "switch(Option('A'), {'X': Value(42), 'Y': Option('Z')})"
    assert (
        repr(switch(Option('A'), {'X': 42, 'Y': Option('Z')}, 43)) ==
        "switch(Option('A'), {'X': Value(42), 'Y': Option('Z')}, Value(43))"
    )

    assert switch('A', {'X': 1})({'A': 'X'}) == 1


def test_case_when():
    c = case(Option('A')).when(
        lambda x: x == 'X',
        42
    ).when(
        lambda x: x == 'Y',
        Option('Z')
    ).otherwise(43)

    assert c({'A': 'X'}) == 42
    assert c({'A': 'Y', 'Z': 44}) == 44
    assert c({'A': 'Z'}) == 43
    with pytest.raises(KeyNotFoundError):
        c.evaluate({})
    with pytest.raises(KeyNotFoundError):
        c.evaluate({'A': 'Y'})

    c.validate({'A': 'X'})
    c.validate({'A': 'Y', 'Z': 44})
    c.validate({'A': 'Z'})
    with pytest.raises(KeyNotFoundError):
        c.validate({})
    with pytest.raises(KeyNotFoundError):
        c.validate({'A': 'Y'})
    c.validate({'A': 'Z'})

    assert c.keys({'A': 'X'}) == {'A'}
    assert c.keys({'A': 'Y', 'Z': 44}) == {'A', 'Z'}
    assert c.keys({'A': 'Z'}) == {'A'}
    with pytest.raises(KeyNotFoundError):
        c.keys({})
    with pytest.raises(KeyNotFoundError):
        c.keys({'A': 'Y'})
    assert c.keys({'A': 'Z'}) == {'A'}

    assert c.explain({'A': 'X'}) == {'A'}
    assert c.explain({'A': 'Y', 'Z': 44}) == {'A', 'Z'}
    assert c.explain({'A': 'Z'}) == {'A'}
    with pytest.raises(InsufficientInformationError):
        c.explain({})
    assert c.explain({'A': 'Y'}) == {'A', 'Z'}
    assert c.explain({'A': 'Z'}) == {'A'}


def test_case_when_no_default():
    c = case(Option('A')).when(
        lambda x: x == 'X',
        42
    ).when(
        lambda x: x == 'Y',
        Option('Z')
    )

    assert c({'A': 'X'}) == 42
    assert c({'A': 'Y', 'Z': 44}) == 44
    with pytest.raises(CaseWhenError):
        c({'A': 'Z'})
    with pytest.raises(KeyNotFoundError):
        c.evaluate({})

    c.validate({'A': 'X'})
    c.validate({'A': 'Y', 'Z': 44})
    with pytest.raises(CaseWhenError):
        c.validate({'A': 'Z'})
    with pytest.raises(KeyNotFoundError):
        c.validate({})

    assert c.keys({'A': 'X'}) == {'A'}
    assert c.keys({'A': 'Y', 'Z': 44}) == {'A', 'Z'}
    with pytest.raises(CaseWhenError):
        c.keys({'A': 'Z'})
    with pytest.raises(KeyNotFoundError):
        c.keys({})

    assert c.explain({'A': 'X'}) == {'A'}
    assert c.explain({'A': 'Y', 'Z': 44}) == {'A', 'Z'}
    with pytest.raises(InsufficientInformationError):
        c.explain({'A': 'Z'})
    with pytest.raises(InsufficientInformationError):
        c.explain({})



import sys

if sys.version_info < (3, 10):
    from typing_extensions import Concatenate, ParamSpec
else:
    from typing import Concatenate, ParamSpec

import inspect
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Set,
    TypeVar,
    cast,
    overload,
)

from .application import PartialApplication
from .types import Evaluatable, MaybeEvaluatable, Options, Transformation, Value

A = TypeVar("A")
B = TypeVar("B", covariant=True)
C = TypeVar("C")
P = ParamSpec("P")


class PipelineStep(Evaluatable[Callable[[A], B]], Transformation[A, B]):
    """A class representing a single step in a pipeline.

    A pipeline step is a single transformation that can be applied to a value.
    Steps can be composed into pipelines using the :code:`+` operator.

    This class should probably not be used directly. Instead, use the :code:`pipeline_step`
    """

    step: Evaluatable[Callable[[A], B]]
    _name: Optional[str]

    def __init__(
        self, step: Evaluatable[Callable[[A], B]], _name: Optional[str] = None
    ) -> None:
        self.step = step
        self._name = _name

    def evaluate(self, options: Options) -> Callable[[A], B]:
        """Evaluate the pipeline step, returning a function that applies the transformation."""
        return self.step.evaluate(options)

    def validate(self, options: Options) -> None:
        """Validate the pipeline step."""
        self.step.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required by the pipeline step."""
        return self.step.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required by the pipeline step."""
        return self.step.explain(options)

    def transform(self, value: A, options: Optional[Options] = None) -> B:
        """Transform a value using the pipeline step.

        Arguments
        ---------
        value : A
            The value to transform.
        options : Optional[Options]
            The options to use when transforming the value.

        Returns
        -------
        B
            The transformed value.
        """
        return self(options)(value)

    def __repr__(self) -> str:
        return f"<PipelineStep {self._name or repr(self.step)}>"

    def __add__(self, other: MaybeEvaluatable[Callable[[B], C]]) -> "Pipeline[A, C]":
        base: Pipeline[A, B] = Pipeline(self)
        return base + other

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, PipelineStep) and self.step == other.step


class _Identity(Generic[A], PipelineStep[A, A]):
    def __new__(cls) -> "_Identity[A]":
        return super().__new__(cls)

    def __init__(self) -> None:
        super().__init__(Value(lambda x: x), "Identity")


Identity: _Identity = _Identity()


class Pipeline(
    Evaluatable[Callable[[A], B]],
    Iterable[Evaluatable[Callable[[Any], Any]]],
    Transformation[A, B],
):
    """A class representing a pipeline of transformations.

    A pipeline is a sequence of transformations that can be applied to a value.
    Pipelines are implemented as a linked list of PipelineSteps, where each step
    is a single transformation that can be applied to a value. Pipelines, like
    PipelineSteps, can be composed using the :code:`+` operator.

    This class should probably not be used directly. Instead, use the :code:`pipeline_step`
    and compose pipelines using the :code:`+` operator.
    """

    tail: PipelineStep[Any, B]
    rest: Optional["Pipeline[A, Any]"]

    @overload
    def __new__(cls) -> "Pipeline[A, A]": ...  # pragma: nocover

    @overload
    def __new__(
        cls, tail: PipelineStep[B, C]
    ) -> "Pipeline[B, C]": ...  # pragma: nocover

    @overload
    def __new__(
        cls, tail: PipelineStep[B, C], rest: "Pipeline[A, B]"
    ) -> "Pipeline[A, C]": ...  # pragma: nocover

    def __new__(
        cls,
        tail: PipelineStep[B, C] = Identity,
        rest: Optional["Pipeline[A, B]"] = None,
    ):
        return super().__new__(cls)

    def __init__(
        self,
        tail: PipelineStep[Any, B] = Identity,
        rest: Optional["Pipeline[A, Any]"] = None,
    ) -> None:
        if rest is not None and rest.empty:
            rest = None

        self.tail = tail
        self.rest = rest

    def evaluate(self, options: Options) -> Callable[[A], B]:
        """Evaluate the pipeline, returning a function that applies all transformations."""
        tail = self.tail.evaluate(options)
        rest = self.rest.evaluate(options) if self.rest else lambda x: x
        return lambda x: tail(rest(x))

    def validate(self, options: Options) -> None:
        """Validate the pipeline."""
        self.tail.validate(options)
        if self.rest:
            self.rest.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required by the pipeline."""
        return self.tail.keys(options) | (
            self.rest.keys(options) if self.rest else set()
        )

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required by the pipeline."""
        return self.tail.explain(options) | (
            self.rest.explain(options) if self.rest else set()
        )

    def transform(self, value: A, options: Optional[Options] = None) -> B:
        """Transform a value using the pipeline."""
        return self(options)(value)

    def __repr__(self) -> str:
        if self.empty:
            return "<Pipeline Identity>"
        return f'<Pipeline {" + ".join(map(repr, self))}>'

    def __add__(self, other: MaybeEvaluatable[Callable[[B], C]]) -> "Pipeline[A, C]":
        if isinstance(other, PipelineStep):
            return Pipeline(other, self)
        elif isinstance(other, Pipeline):
            if other.empty:
                return cast(Pipeline[A, C], self)
            elif other.rest is None:
                return Pipeline(other.tail, self)
            return (self + other.rest) + other.tail
        else:
            return Pipeline(PipelineStep(Evaluatable.ensure(other)), self)

    def __iter__(self) -> Iterator[Evaluatable[Callable[[Any], Any]]]:
        if self.rest is not None:
            yield from self.rest
        yield self.tail

    @property
    def empty(self) -> bool:
        return self.tail == Identity and self.rest is None


def pipeline_step(func: Callable[Concatenate[A, P], B]) -> "PipelineStep[A, B]":
    """Create a pipeline step from a function.

    This is the primary way to create a pipeline step. The function should be used
    as a decorator on the function that will be used as the transformation. The
    function should take at least one argument. The first argument will be the value
    to transform; the remaining arguments will be the parameters for the transformation,
    and should have default values similar to a dataset definition.


    Example Usage
    -------------
    >>> @pipeline_step
    ... def add(x: int, y: int = Option('AMOUNT', 1)) -> int:
    ...     return x + y
    >>>
    >>> @pipeline_step
    ... def multiply(x: int, y: int = Option('FACTOR', 2)) -> int:
    ...     return x * y
    >>>
    >>> pipeline = add + multiply
    >>> pipeline.transform(1)
    4
    >>> pipeline.transform(1, {'AMOUNT': 2, 'FACTOR': 3})
    9
    """
    signature = inspect.signature(func)
    for i, param in enumerate(signature.parameters.values()):
        if i == 0 and param.default is not param.empty:
            raise ValueError(
                "The first parameter of a pipeline step cannot have a default value"
            )
        elif i != 0 and param.default is param.empty:
            raise ValueError(
                "All parameters after the first parameter of a pipeline step "
                "must have a default value"
            )

    return PipelineStep(PartialApplication.lift(func), getattr(func, "__name__", None))



import sys

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

import threading
from typing import Dict, Hashable, Optional, Set, TypeVar

from ._missing import MISSING, MaybeMissing
from .conditional import switch
from .types import Evaluatable, MaybeEvaluatable, Options

A = TypeVar("A", covariant=True)
P = ParamSpec("P")

_MODULE_LOCK = threading.Lock()
_LOCKS: Dict[int, threading.Lock] = {}


def _get_lock(__x: int) -> threading.Lock:
    with _MODULE_LOCK:
        return _LOCKS.setdefault(__x, threading.Lock())


class Overloaded(Evaluatable[A]):
    """A class representing multiple implementations of an Evaluatable.

    This class generalizes the idea of dataset overloads. It takes a dispatch,
    which is an evaluatable (like an Option) that returns a key, and a lookup
    dictionary, which maps keys to evaluatables. When the overloaded object is
    evaluated, it uses the dispatch to determine which evaluatable to use, and
    then evaluates that evaluatable. An optional default evaluatable can be
    provided, which is used if the dispatch key is not found in the lookup
    dictionary, or if the dispatch cannot be evaluated.

    Arguments
    ----------
    dispatch : Evaluatable[Hashable]
        The evaluatable that determines which implementation to use.
    lookup : Dict[Hashable, Evaluatable[A]], optional
        A dictionary mapping keys to evaluatables.
    default : MaybeMissing[MaybeEvaluatable[A]], optional
        The default evaluatable to use if the dispatch key is not found in the
        lookup dictionary. If this is not provided, an exception will be raised
        if the dispatch key is not found.
    """

    dispatch: Evaluatable[Hashable]
    lookup: Dict[Hashable, Evaluatable[A]]
    default: MaybeMissing[Evaluatable[A]]
    _lock: threading.Lock

    def __init__(
        self,
        dispatch: Evaluatable[Hashable],
        lookup: Optional[Dict[Hashable, Evaluatable[A]]],
        default: MaybeMissing[MaybeEvaluatable[A]] = MISSING,
    ):
        self.dispatch = dispatch
        self.lookup = lookup or {}
        self.default = (
            Evaluatable.ensure(default) if default is not MISSING else default
        )
        self._lock = _get_lock(id(self))

    def evaluate(self, options: Options) -> A:
        """Evaluate the dispatch, and then evaluate the selected implementation."""
        return self.switch.evaluate(options)

    def validate(self, options: Options) -> None:
        """Validate the dispatch and the selected implementation."""
        self.switch.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required by the dispatch and the selected implementation.""" ""
        return self.switch.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required by the dispatch and the selected implementation."""
        return self.switch.explain(options)

    def __repr__(self) -> str:
        if self.default is MISSING:
            return f"Overloaded({self.dispatch!r}, {self.lookup!r})"

        return f"Overloaded({self.dispatch!r}, {self.lookup!r}, {self.default!r})"

    def register(self, key: Hashable, value: Evaluatable[A]) -> None:
        """Register a new implementation with the overloaded object.

        Arguments
        ----------
        key : Hashable
            The key for the implementation.
        value : Evaluatable[A]
            The implementation to register.
        """
        with self._lock:
            self.lookup = {**self.lookup, key: value}

    @property
    def switch(self) -> Evaluatable[A]:
        """Return the switch evaluatable that determines which implementation to use."""
        return switch(self.dispatch, self.lookup, default=self.default)

    def __getstate__(self) -> dict:
        return {**self.__dict__, "_lock": id(self)}

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self._lock = _get_lock(state["_lock"])



import hashlib
import json
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import (
    Callable,
    Generic,
    List,
    Mapping,
    Optional,
    Protocol,
    Set,
    TypeVar,
    Union,
    overload,
)

from confectioner.templating import get_dotted_key

from .exceptions import EvaluationError, InsufficientInformationError

JSONScalar = Union[str, int, float, bool, None]
JSON = Union[JSONScalar, Mapping[str, "JSON"], List["JSON"]]
Options = Mapping[str, JSON]


A = TypeVar("A", covariant=True)
B = TypeVar("B", covariant=True)
R = TypeVar("R", covariant=True)
T = TypeVar("T", contravariant=True)
X = TypeVar("X")


class Transformation(Protocol[T, R]):
    """A protocol for objects that can transform values using an options dictionary."""

    def transform(self, value: T, options: Optional[Options] = None) -> R:
        raise NotImplementedError  # pragma: nocover


class Validatable(ABC):
    """Abstract base class for objects that can be validated against an options dictionary."""

    @abstractmethod
    def validate(self, options: Options) -> None:
        """Validate the object.

        Arguments
        ----------
        options : Options
            The options dictionary to validate against.

        Raises
        ------
        KeyNotFoundError
            If a required key is not found in the options dictionary.

        Notes
        -----
        This method should be called recursively on all child objects where
        applicable.
        """
        raise NotImplementedError  # pragma: nocover


class Cacheable(ABC):
    """Abstract base class for objects that can be cached."""

    @abstractmethod
    def keys(self, options: Options) -> Set[str]:
        """
        Return the keys from the Options that this object depends on.

        This should only return the keys that are present in the options dictionary.
        If a key is not present but is required, a KeyNotFoundError should be raised.

        Arguments
        ----------
        options : Options
            The options dictionary to validate against.

        Returns
        -------
        Set[str]
            The keys that this object depends on.

        Raises
        ------
        KeyNotFoundError
            If a required key is not found in the options dictionary.

        Notes
        -----
        This method should be called recursively on all child objects where
        applicable.
        """
        raise NotImplementedError  # pragma: nocover

    def fingerprint(self, options: Options) -> bytes:
        """Return a fingerprint, which is a unique identifier for a given evaluation."""
        fingerprint = hashlib.blake2b(digest_size=64)

        for key in sorted(self.keys(options)):
            fingerprint.update(key.encode())
            fingerprint.update(
                json.dumps(get_dotted_key(key, options), sort_keys=True).encode()
            )

        return fingerprint.digest()


class Explainable(ABC):
    """Abstract base class for objects that can explain themselves."""

    @abstractmethod
    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return all keys that this object depends on.

        This should return all keys that this object depends on, including those that are not
        present in the options dictionary. If the keys required cannot be determined, an
        InsufficientInformationError should be raised.

        Arguments
        ----------
        options : Options
            The options dictionary to validate against.

        Returns
        -------
        Set[str]
            The keys that this object depends on.

        Raises
        ------
        InsufficientInformationError
            If the keys required cannot be determined.

        Notes
        -----
        This method should be called recursively on all child objects where
        applicable.
        """


class Evaluatable(Generic[A], Cacheable, Explainable, Validatable, ABC):
    """Abstract base class for objects that can be evaluated with an Options dictionary.

    This ABC is used to define a common interface for objects that can be
    evaluated using an options dictionary. Datasets, Options, and other types
    defined in this library implement subclass this ABC. This allows for
    polymorphic behavior when evaluating objects, and allows for third-party
    extensions to be created that can be used within the labrea framework.
    """

    def __call__(self, options: Optional[Options] = None) -> A:
        """Evaluate the object.

        An alias for the evaluate method, but can be called with no arguments.

        Arguments
        ----------
        options : Optional[Options]
            The options dictionary to evaluate against. An empty dictionary is
            used if no options are provided

        Returns
        -------
        A
            The evaluated value.

        Raises
        ------
        KeyNotFoundError
            If a required key is not found in the options dictionary.
        """
        return self.evaluate(options or {})

    @abstractmethod
    def evaluate(self, options: Options) -> A:
        """Evaluate the object.

        Arguments
        ----------
        options : Options
            The options dictionary to evaluate against.

        Returns
        -------
        A
            The evaluated value.

        Raises
        ------
        EvaluationError
            If the object cannot be evaluated.

        Notes
        -----
        This method should be called recursively on all child objects where
        applicable.
        """
        raise NotImplementedError  # pragma: nocover

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the object.

        Returns
        -------
        str
            A string representation of the object.
        """
        raise NotImplementedError  # pragma: nocover

    @staticmethod
    def unit(value: T) -> "Value[T]":
        """Wrap a value in a Value object.

        This method is used to wrap a value in a Value object. This is useful
        when you want to treat a value as an Evaluatable.

        Arguments
        ----------
        value : A
            The value to wrap.

        Returns
        -------
        Value[A]
            The wrapped value.
        """
        return Value(value)

    @overload
    @staticmethod
    def ensure(value: "Evaluatable[T]") -> "Evaluatable[T]": ...

    @overload
    @staticmethod
    def ensure(value: T) -> "Evaluatable[T]": ...

    @staticmethod
    def ensure(value: "MaybeEvaluatable[T]") -> "Evaluatable[T]":
        """Ensure that a value is an Evaluatable.

        This method is used to ensure that a value is an Evaluatable. If the
        value is already an Evaluatable, it is returned as is. If the value is
        not an Evaluatable, it is wrapped in a Value object.

        Arguments
        ----------
        value : MaybeEvaluatable[A]
            The value to ensure is an Evaluatable.

        Returns
        -------
        Evaluatable[A]
            The value as an Evaluatable.
        """
        if isinstance(value, Evaluatable):
            return value
        return Value(value)

    def apply(self, func: "MaybeEvaluatable[Callable[[A], B]]") -> "Evaluatable[B]":
        """Lazy application of a function to the result of evaluating the object.

        This method is used to apply a function to the result of evaluating the
        object. The function is not evaluated until the object is evaluated.
        Equivalently can use the :code:`>>` operator.

        Arguments
        ----------
        func : MaybeEvaluatable[Callable[[A], B]]
            The function to apply to the result of evaluating the object. Either a
            function of one argument, or an Evaluatable that evaluates to a function
            of one argument.

        Returns
        -------
        Evaluatable[B]
            A new evaluatable, that when evaluated, applies the function to the
            result of evaluating the source object.


        Example Usage
        -------------
        >>> from labrea import Option
        >>>
        >>> Option('A').apply(str.upper)({'A': 'foo'})
        'FOO'
        >>>
        >>> (Option('B') >> str.upper)({'B': 'bar'})
        'BAR'
        """
        if not isinstance(func, Evaluatable) and not callable(func):
            raise TypeError(f"Cannot apply object of type ({type(func)})")

        return Apply(self, self.ensure(func))

    def bind(self, func: Callable[[A], "Evaluatable[B]"]) -> "Evaluatable[B]":
        """Lazy bind a function to the result of evaluating the object.

        This method is used to bind a function to the result of evaluating the
        object. The function should take the result of evaluating the object,
        and return an Evaluatable. This is useful for when you want to use the
        result of evaluating the object to determine the next step in the
        evaluation process.

        Arguments
        ----------
        func : Callable[[A], Evaluatable[B]]
            The function to bind to the result of evaluating the object.

        Returns
        -------
        Evaluatable[B]
            A new evaluatable, that when evaluated, binds the function to the
            result of evaluating the source object.

        Example Usage
        -------------
        >>> from labrea import Option
        >>>
        >>> x_if_a_neg_else_y = Option('A').bind(lambda a: Option('X') if a < 0 else Option('Y'))
        >>> x_if_a_neg_else_y({'A': -1, 'X': 'foo', 'Y': 'bar'})
        'foo'
        >>> x_if_a_neg_else_y({'A': 1, 'X': 'foo', 'Y': 'bar'})
        'bar'
        """
        if not callable(func):
            raise TypeError(f"Cannot bind object of type ({type(func)})")
        return Bind(self, func)

    @property
    def result(self) -> A:
        """Dummy property used to appease type checkers.

        When creating a dataset, type checkers will complain that the default
        values are not of the correct type. This property is used to tell the
        type checker that the default value is of the correct type.
        """
        return self  # type: ignore

    def __rshift__(
        self, other: "MaybeEvaluatable[Callable[[A], B]]"
    ) -> "Evaluatable[B]":
        return self.apply(other)


class Value(Evaluatable[A]):
    """Simple wrapper for a plain value.

    This class is used to wrap a value that is not an Evaluatable and make it
    an Evaluatable.

    Arguments
    ----------
    value : A
        The key to wrap.
    """

    value: A

    def __init__(self, value: A) -> None:
        self.value = value

    def evaluate(self, options: Options) -> A:
        """Return the wrapped value."""
        try:
            return deepcopy(self.value)
        except Exception:  # noqa: E722
            return self.value

    def validate(self, options: Options) -> None:
        """Always passes validation."""
        pass

    def keys(self, options: Options) -> Set[str]:
        """Return an empty set, as this object does not depend on any keys."""
        return set()

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return an empty set, as this object does not depend on any keys."""
        return set()

    def __repr__(self) -> str:
        return f"Value({self.value!r})"

    def __eq__(self, other):
        return isinstance(other, Value) and self.value == other.value


class Apply(Generic[A, B], Evaluatable[B]):
    """A class representing the application of a function to the result of evaluating an object.

    This class is used to apply a function to the result of evaluating an object.
    Apply objects are not usually created directly, instead the :code:`apply` method
    is used on an Evaluatable object.
    """

    evaluatable: Evaluatable[A]
    func: Evaluatable[Callable[[A], B]]

    def __init__(
        self, evaluatable: Evaluatable[A], func: Evaluatable[Callable[[A], B]]
    ) -> None:
        self.evaluatable = evaluatable
        self.func = func

    def evaluate(self, options: Options) -> B:
        """Apply the function to the result of evaluating the object."""
        value = self.evaluatable(options)
        return self.func(options)(value)

    def validate(self, options: Options) -> None:
        """Validate the source object and the function to apply to it."""
        self.evaluatable.validate(options)
        self.func.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the keys the source object and the function depend on."""
        return self.evaluatable.keys(options) | self.func.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the keys the source object and the function depend on."""
        return self.evaluatable.explain(options) | self.func.explain(options)

    def __repr__(self) -> str:
        return f"{self.evaluatable!r}.apply({self.func!r})"


class Bind(Generic[A, B], Evaluatable[B]):
    """A class representing the binding of a function to the result of evaluating an object.

    This class is used to bind a function to the result of evaluating an object.
    Bind objects are not usually created directly, instead the :code:`bind` method
    is used on an Evaluatable object.
    """

    evaluatable: Evaluatable[A]
    func: Callable[[A], Evaluatable[B]]

    def __init__(
        self, evaluatable: Evaluatable[A], func: Callable[[A], Evaluatable[B]]
    ) -> None:
        self.evaluatable = evaluatable
        self.func = func

    def evaluate(self, options: Options) -> B:
        """Bind the function to the result of evaluating the object."""
        return self.func(self.evaluatable(options)).evaluate(options)

    def validate(self, options: Options) -> None:
        """Validate the source object and the function"""
        self.evaluatable.validate(options)
        self.func(self.evaluatable(options)).validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the keys the source object, function, and result depend on.

        Because the function is bound to the result of evaluating the source object,
        the source object must be evaluated to determine the keys that the function's
        result depends on.
        """
        return self.evaluatable.keys(options) | self.func(
            self.evaluatable(options)
        ).keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the keys the source object, function, and result depend on.

        Because the function is bound to the result of evaluating the source object,
        the source object must be evaluated to determine the keys that the function's
        result depends on.
        """
        try:
            return self.evaluatable.explain(options) | self.func(
                self.evaluatable(options)
            ).explain(options)
        except EvaluationError as e:
            raise InsufficientInformationError(f"Cannot explain {self}", self) from e

    def __repr__(self) -> str:
        return f"{self.evaluatable!r}.bind({self.func!r})"


MaybeEvaluatable = Union[Evaluatable[X], X]



from abc import ABC, abstractmethod
from typing import Callable, List, Optional, Set, TypeVar

from .option import Option
from .types import (
    Evaluatable,
    Explainable,
    MaybeEvaluatable,
    Options,
    Transformation,
    Validatable,
)

A = TypeVar("A")

_EFFECTS_DISABLED = Option("LABREA.EFFECTS.DISABLED", False)


class Effect(Transformation[A, None], Validatable, Explainable, ABC):
    """Abstract base class for effects.

    Effects are transformations that are applied to values, but do not return a new value.
    They are intended to encapsulate side effects, such as logging, data validation, or
    other operations that do not affect the value being returned by an Evaluatable.
    """

    @abstractmethod
    def transform(self, value: A, options: Optional[Options] = None) -> None:
        """Apply the effect to a value.

        Arguments
        ----------
        value : A
            The value to apply the effect to.
        options : Options
            The options dictionary to evaluate against.

        Returns
        -------
        A
            The value after the effect has been applied.

        Raises
        ------
        EvaluationError
            If the effect cannot be applied to the value.
        """
        raise NotImplementedError  # pragma: nocover


class ChainedEffect(Effect[A]):
    """An effect that chains multiple effects together.

    This effect applies a sequence of effects to a value in order.

    Arguments
    ---------
    *effects : Effect
        The effects to chain together.
    """

    effects: List[Effect[A]]

    def __init__(self, *effects: Effect[A]):
        self.effects = list(effects)

    def transform(self, value: A, options: Optional[Options] = None) -> None:
        """Perform each effect in sequence."""
        for effect in self.effects:
            effect.transform(value, options)

    def validate(self, options: Options) -> None:
        """Validate each effect."""
        for effect in self.effects:
            effect.validate(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required to perform each effect."""
        return set().union(*(effect.explain(options) for effect in self.effects))

    def __repr__(self) -> str:
        return f"ChainedEffect({', '.join(map(repr, self.effects))})"


class CallbackEffect(Effect[A]):
    """An effect that applies a callback to a value.

    This allows any function of one value to be used as an effect. The
    callback can also be an Evaluatable that returns a function,
    which allows :code:`Pipeline` objects to be used as effects.

    Arguments
    ---------
    callback : MaybeEvaluatable[Callable[[A], None]]
        The callback to apply to the value. Either a function of one
        value that returns None, or an Evaluatable that returns such
        a function.
    """

    callback: Evaluatable[Callable[[A], None]]

    def __init__(self, callback: MaybeEvaluatable[Callable[[A], None]]):
        self.callback = Evaluatable.ensure(callback)

    def transform(self, value: A, options: Optional[Options] = None) -> None:
        """Apply the callback to the value."""
        self.callback(options)(value)

    def validate(self, options: Options) -> None:
        """Validate the callback."""
        self.callback.validate(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required to perform the callback."""
        return self.callback.explain(options)

    def __repr__(self) -> str:
        return f"CallbackEffect({self.callback!r})"


class Computation(Evaluatable[A]):
    """A computation that applies an effect to a value.

    This class combines an Evaluatable with an Effect to create a new
    Evaluatable that applies the effect to the value returned by the
    original Evaluatable. This allows side effects to be applied to
    values without modifying the original Evaluatable.

    If the :code:`LABREA.EFFECTS.DISABLED` option is set to True, the
    effect will not be applied to the value. This allows the effect to
    be disabled in certain contexts, such as testing or debugging.


    Arguments
    ---------
    evaluatable : Evaluatable[A]
        The Evaluatable to apply the effect to.
    effect : Effect[A]
        The Effect to apply to the value returned by the Evaluatable.
    """

    evaluatable: Evaluatable[A]
    effect: Effect[A]

    def __init__(self, evaluatable: Evaluatable[A], effect: Effect[A]):
        self.evaluatable = evaluatable
        self.effect = effect

    def evaluate(self, options: Options) -> A:
        """Evaluate the Evaluatable and apply the Effect to the result before returning."""
        value = self.evaluatable.evaluate(options)

        if not _EFFECTS_DISABLED(options):
            self.effect.transform(value, options)

        return value

    def validate(self, options: Options) -> None:
        """Validate the Evaluatable and the Effect."""
        self.evaluatable.validate(options)

        if not _EFFECTS_DISABLED(options):
            self.effect.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required to evaluate the Evaluatable."""
        return self.evaluatable.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required to evaluate the Evaluatable and apply the Effect."""
        return (
            self.evaluatable.explain(options)
            if _EFFECTS_DISABLED(options)
            else self.evaluatable.explain(options) | self.effect.explain(options)
        )

    def __repr__(self) -> str:
        return f"Computation({self.evaluatable!r}, {self.effect!r})"
